from typing import Any

class CyclicGraphException(Exception):
    """Raised when a cycle is detected in a graph which doesn't allow one."""

    def __init__(self, nodes: set[Any], *args: object) -> None:
        """Initializes a new instance.

        :param nodes: The list of nodes that form the cycle.
        :type nodes: set[Any]
        """

        super().__init__(*args)
        self.__nodes: set[Any] = nodes

    @property
    def nodes(self) -> set[Any]:
        """Gets the list of nodes that form the cycle.

        :return: The list of nodes that form the cycle.
        :rtype: set[Any]
        """

        return self.__nodes



from typing import Any

class DuplicateEdgeException(Exception):
    """Raised when a duplicate edge is detected in a graph that doesn't allow them."""

    def __init__(self, node1: Any, node2: Any, *args: object) -> None:
        """Initializes a new instance.

        :param node1: The first of the nodes with the duplicate edges.
        :type node1: Any
        :param node2: The second of the nodes with the duplicate edges.
        :type node2: Any
        """

        super().__init__(*args)
        self.__node1: Any = node1
        self.__node2: Any = node2
    
    @property
    def node1(self) -> Any:
        """Gets the first of the two nodes that have the duplicate edges.

        :return: The first of the two nodes that have the duplicate edges.
        :rtype: Any
        """

        return self.__node1
    
    @property
    def node2(self) -> Any:
        """Gets the second of the two nodes that have the duplicate edges.

        :return: The second of the two nodes that have the duplicate edges.
        :rtype: Any
        """

        return self.__node2



"""Graph specific exception types."""

from .cyclic_graph_exception import CyclicGraphException
from .disconnected_sub_graph_exception import DisconnectedSubGraphException
from .duplicate_edge_exception import DuplicateEdgeException
from .duplicate_node_exception import DuplicateNodeException



from dataclasses import dataclass
from typing import Any

from .injectable_registration import InjectableRegistration

@dataclass(kw_only=True)
class InjectableInstanceRegistration(InjectableRegistration):
    """Holds information about the registration of an injectable instance."""

    injectable_instance: Any
    """Gets or sets the injectable instance."""



from dataclasses import dataclass

from .injectable_registration import InjectableRegistration
from .injectable_scope_type import InjectableScopeType

@dataclass(kw_only=True)
class InjectableTypeRegistration(InjectableRegistration):
    """Holds information about the registration of an injectable type."""

    injectable_type: type
    """Gets or sets the type of the injectable object."""

    is_generic: bool = False
    """Gets or sets whether the injectable should be constructed
    by passing the generic type arguments of the contract
    as generic type parameters to the injectable.

    This behavior will result in a runtime error if the specified type
    is not a generic type or there is a signature mismatch between
    the injectable type and the contract type.
    """

    scope: InjectableScopeType = InjectableScopeType.TRANSIENT
    """Gets or sets the lifetime scope type of the created instances."""



from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import Any, get_args

from kanata.exceptions import DependencyResolutionException
from kanata.models import (
    ClosedGenericTypeId, ClosedGenericTypeInfo, IInstanceCollection, InjectableInstanceRegistration,
    InjectableRegistration, InjectableScopeType, InjectableTypeRegistration
)
from kanata.utils import get_dependent_contracts
from .iresolver import IResolver
from .resolver_context import ResolverContext

_SCOPE_TYPE_RANKS: dict[InjectableScopeType, int] = {
    InjectableScopeType.TRANSIENT: 0,
    InjectableScopeType.SCOPED: 1,
    InjectableScopeType.SINGLETON: 2
}

class ResolverBase(ABC, IResolver):
    """Abstract base class for a resolver."""

    @abstractmethod
    def resolve(
        self,
        context: ResolverContext,
        registration: InjectableRegistration,
        injectable_type: type
    ) -> Any:
        ...

    @staticmethod
    def _is_captive_dependency(
        dependee_scope: InjectableScopeType,
        dependent_scope: InjectableScopeType
    ) -> bool:
        """Determines if the dependee and dependent scopes
        result in a captive dependency.

        :param dependee_scope: The scope of the dependee.
        :type dependee_scope: InjectableScopeType
        :param dependent_scope: The scope of the dependent.
        :type dependent_scope: InjectableScopeType
        :return: True, if a captive dependency is detected.
        :rtype: bool
        """

        ranked_scopes = sorted(
            (dependee_scope, dependent_scope),
            key=lambda i: _SCOPE_TYPE_RANKS[i]
        )
        return ranked_scopes[0] != dependee_scope

    def _on_captive_dependency_detected(
        self,
        injectable: type,
        contract: type
    ) -> None:
        """Invoked when a captive dependency is detected
        so that the resolver can handle the situation.

        This method should be overridden by subclasses
        and it does nothing by default.

        :param injectable: The type of the causing injectable.
        :type injectable: type
        :param contract: The type of the dependent contract.
        :type contract: type
        """

    def _get_dependencies(
        self,
        context: ResolverContext,
        injectable: type,
        dependee_scope: InjectableScopeType
    ) -> list[Any]:
        """Gets the dependencies of the specified injectable.

        :param context: Contextual information for the resolver.
        :type context: ResolverContext
        :param injectable: The type of the injectable for which to get the dependencies.
        :type injectable: type
        :param dependee_scope: The scope of the dependee.
        :type dependee_scope: InjectableScopeType
        :raises DependencyResolutionException: Raised when a dependency cannot be satisfied.
        :return: The list of dependencies.
        :rtype: list[Any]
        """

        dependent_contracts = get_dependent_contracts(injectable)
        dependent_injectables = list[Any]()
        for dependent_contract, is_multi in dependent_contracts:
            candidate_instances = self.__get_candidate_dependent_instances(
                context,
                injectable,
                dependee_scope,
                dependent_contract
            )

            if is_multi:
                dependent_injectables.append(tuple(candidate_instances))
            elif len(candidate_instances) > 0:
                dependent_injectables.append(candidate_instances[0])
            else:
                raise DependencyResolutionException(
                    injectable,
                    (
                        "Cannot satisfy the dependency"
                        f" of '{injectable}' on '{dependent_contract}'."
                    )
                )

        return dependent_injectables

    @staticmethod
    def __get_candidates_by_type_registration(
        closed_generic_types: dict[ClosedGenericTypeId, ClosedGenericTypeInfo],
        instances: IInstanceCollection,
        registration: InjectableTypeRegistration,
        dependent_contract: type
    ) -> Iterable[Any]:
        return (
            ResolverBase.__get_closed_generic_instances(
                closed_generic_types,
                instances,
                registration,
                dependent_contract
            )
            if registration.is_generic
            else instances.get_instances_by_injectable(
                registration.injectable_type,
                registration.scope
            )
        )

    @staticmethod
    def __get_closed_generic_instances(
        closed_generic_types: dict[ClosedGenericTypeId, ClosedGenericTypeInfo],
        instances: IInstanceCollection,
        origin_registration: InjectableTypeRegistration,
        contract_type: type
    ) -> Iterable[Any]:
        type_argument = get_args(contract_type)[0]
        generic_type_id = ClosedGenericTypeId(origin_registration.injectable_type, type_argument)
        if not (closed_generic_type_info := closed_generic_types.get(generic_type_id)):
            return ()

        return instances.get_instances_by_injectable(
            closed_generic_type_info.closed_generic_type,
            origin_registration.scope
        )

    def __get_candidate_dependent_instances(
        self,
        context: ResolverContext,
        injectable: type,
        dependee_scope: InjectableScopeType,
        dependent_contract: type
    ) -> list[Any]:
        registrations = context.catalog.get_registrations_by_contract(dependent_contract)
        candidate_instances = []
        for registration in registrations:
            if isinstance(registration, InjectableTypeRegistration):
                if ResolverBase._is_captive_dependency(dependee_scope, registration.scope):
                    self._on_captive_dependency_detected(injectable, dependent_contract)

                candidate_instances.extend(
                    ResolverBase.__get_candidates_by_type_registration(
                        context.closed_generic_types,
                        context.instances,
                        registration,
                        dependent_contract
                    )
                )
            elif isinstance(registration, InjectableInstanceRegistration):
                candidate_instances.append(registration.injectable_instance)
            else: raise DependencyResolutionException(
                type(registration),
                "Unsupported type of injectable registration."
            )

        return candidate_instances



from typing import Any, TypeVar

import structlog

from kanata.constants import LOGGER_NAME
from kanata.exceptions import DependencyResolutionException
from kanata.models import (
    InjectableInstanceRegistration, InjectableRegistration, InjectableScopeType,
    InjectableTypeRegistration
)
from .default_resolver_options import DefaultResolverOptions
from .resolver_base import ResolverBase
from .resolver_context import ResolverContext

TInjectable = TypeVar("TInjectable")

class DefaultResolver(ResolverBase):
    """Default implementation of a resolver."""

    def __init__(
        self,
        options: DefaultResolverOptions | None = None
    ) -> None:
        super().__init__()
        self.__options = options or DefaultResolverOptions()
        self.__log = structlog.get_logger(logger_name=LOGGER_NAME)

    def resolve(
        self,
        context: ResolverContext,
        registration: InjectableRegistration,
        injectable_type: type
    ) -> Any:
        if isinstance(registration, InjectableInstanceRegistration):
            return registration.injectable_instance
        if not isinstance(registration, InjectableTypeRegistration):
            raise DependencyResolutionException(
                type(registration),
                "Unsupported type of injectable registration."
            )

        # For singleton and scoped injectables, we know that there exists one and only one
        # instance for all of the associated contracts, therefore we can find and return
        # that specific one if it is created already.
        if (
            registration.scope in (InjectableScopeType.SINGLETON, InjectableScopeType.SCOPED)
            and (matching_instances := context.instances.get_instances_by_injectable(
                injectable_type,
                registration.scope
            ))
            and (instance := next(iter(matching_instances), None))
        ):
            return instance

        return injectable_type(
            *self._get_dependencies(
                context,
                injectable_type,
                registration.scope
            )
        )

    def _on_captive_dependency_detected(
        self,
        injectable: type,
        contract: type
    ) -> None:
        # Unless turned off, issue a warning, because a captive dependency
        # may be the result of a coding error, as it's generally undesirable.
        error_message = (
            "Detected captive dependency."
            f" Singleton '{injectable}' depends on transient '{contract}'."
        )
        if not self.__options.suppress_captive_dependency_warnings:
            self.__log.warn(error_message)
        if self.__options.raise_on_captive_dependency:
            raise DependencyResolutionException(injectable, error_message)



from dataclasses import dataclass, field

from .closed_generic_type_id import ClosedGenericTypeId
from .injectable_type_registration import InjectableTypeRegistration

@dataclass(kw_only=True)
class ClosedGenericTypeInfo:
    """Information about a dynamically created closed generic type."""

    closed_generic_type: type
    """The dynamically created closed generic type."""

    generic_type_argument: type
    """The generic type argument of the closed generic type."""

    origin_registration: InjectableTypeRegistration
    """The registration for which the closed generic type was created."""

    identifier: ClosedGenericTypeId = field(init=False)
    """The identifier of the closed generic type."""

    def __post_init__(self) -> None:
        self.identifier = ClosedGenericTypeId(
            self.origin_registration.injectable_type,
            self.generic_type_argument
        )



"""Library-wide exception types."""

from .argument_exception import ArgumentException
from .dependency_resolution_exception import DependencyResolutionException
from .injectable_registration_exception import InjectableRegistrationException



"""Utility methods for dictionaries."""

from collections.abc import Callable
from typing import TypeVar

TKey = TypeVar("TKey")
TValue = TypeVar("TValue")

def get_or_add(
    dictionary: dict[TKey, TValue],
    key: TKey,
    value_factory: Callable[[TKey], TValue]
) -> TValue:
    """Gets the value associated to the specified key from the specified dictionary,
    or adds the specified value if the key cannot be found.

    :param dictionary: The dictionary to use.
    :type dictionary: dict[TKey, TValue]
    :param key: The key to look for.
    :type key: TKey
    :param value_factory: The value to add when the key cannot be found.
    :type value_factory: Callable[[TKey], TValue]
    :return: If exists, the value currently present in the dictionary; otherwise, the newly added value.
    :rtype: TValue
    """

    existing_value = dictionary.get(key)
    if not existing_value:
        dictionary[key] = existing_value = value_factory(key)
    return existing_value



from collections.abc import Generator
from typing import Any

from .iinstance_collection import IInstanceCollection
from .injectable_scope_type import InjectableScopeType

class InstanceCollection(IInstanceCollection):
    """A container for resolved instances."""

    def __init__(self) -> None:
        self.__instances = dict[InjectableScopeType, dict[type, set]]()

    def get_instances_by_injectable(
        self,
        injectable_type: type,
        scope_type: InjectableScopeType | None = None
    ) -> Generator[Any, None, None]:
        if scope_type is None:
            for injectables_by_contract in self.__instances.values():
                yield from injectables_by_contract.get(injectable_type, ())
            return

        if not (injectables_by_contract := self.__instances.get(scope_type)):
            return

        yield from injectables_by_contract.get(injectable_type, ())

    def add_instance(
        self,
        scope_type: InjectableScopeType,
        injectable_type: type,
        instance: Any
    ) -> None:
        """Adds the specified instance to the collection associated to
        the specified injectable type and scope.

        :param scope_type: The scope of the injectable.
        :type scope_type: InjectableScopeType
        :param injectable_type: The type of the injectable.
        :type injectable_type: type
        :param instance: The injectable instance to be added.
        :type instance: Any
        """

        if not (instances_by_scope := self.__instances.get(scope_type)):
            self.__instances[scope_type] = instances_by_scope = {}
        if not (instances := instances_by_scope.get(injectable_type)):
            instances_by_scope[injectable_type] = instances = set()
        instances.add(instance)



from typing import Any, Protocol

from kanata.models import InjectableRegistration
from .resolver_context import ResolverContext

class IResolver(Protocol):
    """Interface for a service that can resolve instances of injectables."""

    def resolve(
        self,
        context: ResolverContext,
        registration: InjectableRegistration,
        injectable_type: type
    ) -> Any:
        """Resolves an instance of a specific injectable.

        :param context: Contextual information for the resolver.
        :type context: ResolverContext
        :param injectable_type: The injectable that needs to be resolved. For generic injectables, this is the closed generic type.
        :type injectable_type: type
        :return: The resolved injectable instance.
        :rtype: TInjectable
        """
        ...



from collections.abc import Generator

from kanata.exceptions import ArgumentException
from .edge import Edge
from .graph import Graph
from .tnode import TNode

class BidirectedGraph(Graph[TNode]):
    """A type of graph whose edges specify a direction between nodes."""

    def __init__(self) -> None:
        super().__init__()

    def get_out_edges(self, node: TNode) -> Generator[Edge[TNode], None, None]:
        """Gets the edges that point outward from the specified node.

        :param node: The node for which to get the out-edges.
        :type node: TNode
        :raises ArgumentException: Raised when the specified node is not in the graph.
        :yield: The edges that point outward from the specified node.
        :rtype: Generator[Edge[TNode], None, None]
        """

        if node not in self:
            raise ArgumentException("node", node, "The specified node is not in the graph.")

        for edge in self.edges:
            if edge.source == node:
                yield edge

    def _is_same_edge(self, existing_edge: Edge, new_edge: Edge) -> bool:
        return existing_edge == new_edge



from __future__ import annotations

from typing import Protocol, TypeVar

TInjectable = TypeVar("TInjectable")

class ILifetimeScope(Protocol):
    """Interface for an injectable lifetime scope
    that manages the lifetimes of injectables
    and provides access to them.
    """

    def resolve(self, injectable: type[TInjectable]) -> TInjectable:
        """Resolves an instance of an injectable associated to the specified interface.

        :param injectable: The injectable for which to get a resolved instance.
        :type injectable: type[TInjectable]
        :raises DependencyResolutionException: Raised when dependency resolution fails.
        :return: An instance of the appropriate injectable.
        :rtype: TInjectable
        """
        ...

    def create_child_scope(self) -> ILifetimeScope:
        """Creates a lifetime scoped attached to the current one.
        This child will have access to the same injectable catalog,
        as well as singleton instances. Scoped instances will die
        along with this child.

        :return: A lifetime scope attached to the current lifetime scope.
        :rtype: ILifetimeScope
        """
        ...



import types
from collections.abc import Iterable
from typing import Any, get_args

import structlog

from kanata.catalogs import IInjectableCatalog
from kanata.utils import get_dependent_contracts
from .constants import LOGGER_NAME
from .exceptions import DependencyResolutionException
from .graphs import BidirectedGraph
from .graphs.sorting import topological_sort
from .ilifetime_scope import ILifetimeScope, TInjectable
from .models import (
    ClosedGenericTypeId, ClosedGenericTypeInfo, InjectableInstanceRegistration,
    InjectableRegistration, InjectableScopeType, InjectableTypeRegistration, InstanceCollection
)
from .resolvers import DefaultResolver, IResolver, ResolverContext

class LifetimeScope(ILifetimeScope):
    """An injectable lifetime scope
    that manages the lifetimes of injectables
    and provides access to them."""

    def __init__(
        self,
        catalog: IInjectableCatalog,
        resolvers: tuple[IResolver, ...] | None = None,
        _parent: ILifetimeScope | None = None
    ) -> None:
        self.__catalog = catalog
        self.__resolvers: tuple[IResolver, ...] = resolvers or (DefaultResolver(),)
        self.__parent = _parent
        self.__log = structlog.get_logger(logger_name=LOGGER_NAME)
        self.__instances = InstanceCollection()
        # The below dictionaries are used for tracking
        # the dynamically created closed generic types.
        self.__closed_generic_type_infos_by_id = dict[ClosedGenericTypeId, ClosedGenericTypeInfo]()
        self.__closed_generic_type_infos_by_type = dict[type, ClosedGenericTypeInfo]()

    def resolve(self, injectable: type[TInjectable]) -> TInjectable:
        if self.__parent and self.__should_resolve_via_parent(injectable):
            return self.__parent.resolve(injectable)

        dependency_graph = self.__build_dependency_graph_for(injectable)
        resolver_context = ResolverContext(
            catalog=self.__catalog,
            closed_generic_types=self.__closed_generic_type_infos_by_id,
            instances=self.__instances
        )
        instance = None
        for current_injectable in topological_sort(dependency_graph, injectable):
            self.__log.debug("Resolving injectable", type=current_injectable)
            instance = self.__resolve_injectable(resolver_context, current_injectable)
            self.__log.debug("Resolved injectable", type=current_injectable)

        if not isinstance(instance, injectable):
            raise DependencyResolutionException(
                type(instance),
                (
                    "Got an unexpected type during dependency resolution."
                    " This is likely an error in the algorithm."
                )
            )

        return instance

    def create_child_scope(self) -> ILifetimeScope:
        return LifetimeScope(self.__catalog, self.__resolvers, _parent=self)

    @staticmethod
    def __get_injectable_scope_type(registration: InjectableRegistration) -> InjectableScopeType:
        match registration:
            case InjectableTypeRegistration():
                return registration.scope
            case InjectableInstanceRegistration():
                return InjectableScopeType.SINGLETON
            case _:
                raise DependencyResolutionException(
                    type(registration),
                    "Unsupported type of injectable registration."
                )

    def __build_dependency_graph_for(self, injectable: type) -> BidirectedGraph[type]:
        graph: BidirectedGraph[type] = BidirectedGraph()
        injectables_to_resolve: list[type] = [injectable]
        while injectables_to_resolve:
            dependee_injectable = injectables_to_resolve.pop()
            if not graph.try_add_node(dependee_injectable):
                # This type's dependency chain has already been mapped.
                continue

            self.__log.debug("Gathering dependent contracts", dependee=dependee_injectable)
            dependent_contracts = get_dependent_contracts(dependee_injectable)
            for dependent_contract, is_multi in dependent_contracts:
                self.__log.debug(
                    "Found dependent contract",
                    dependee=dependee_injectable,
                    dependent=dependent_contract,
                    is_multi=is_multi
                )
                dependent_registrations = self.__catalog.get_registrations_by_contract(
                    dependent_contract
                )
                if len(dependent_registrations) == 0 and not is_multi:
                    raise DependencyResolutionException(
                        dependee_injectable,
                        "Cannot satisfy the dependency"
                        f" of {dependee_injectable} on {dependent_contract}."
                    )

                injectables_to_resolve.extend(
                    self.__mark_dependent_types(
                        graph,
                        dependee_injectable,
                        dependent_contract,
                        dependent_registrations
                    )
                )

        return graph

    def __resolve_injectable(
        self,
        resolver_context: ResolverContext,
        injectable: type
    ) -> Any:
        if closed_generic_type_info := self.__closed_generic_type_infos_by_type.get(injectable):
            registration = closed_generic_type_info.origin_registration
        else:
            registration = self.__catalog.get_registration_by_injectable(injectable)
            if not registration:
                raise DependencyResolutionException(
                    injectable,
                    f"Cannot find the registration for injectable '{injectable}'."
                    " It is possible that this type is not an injectable."
                )

        for resolver in self.__resolvers:
            if not (instance := resolver.resolve(resolver_context, registration, injectable)):
                continue

            self.__log.debug("Instantiated injectable", injectable=injectable)
            scope = LifetimeScope.__get_injectable_scope_type(registration)
            self.__instances.add_instance(scope, injectable, instance)

            return instance

        raise DependencyResolutionException(
            injectable,
            "None of the resolvers could resolve an instance of the specified type."
        )

    def __should_resolve_via_parent(
        self,
        injectable: type
    ) -> bool:
        registration = self.__catalog.get_registration_by_injectable(injectable)
        match registration:
            case InjectableTypeRegistration(scope=InjectableScopeType.SINGLETON): return True
            case InjectableInstanceRegistration(): return True
            case _: return False

    def __mark_dependent_types(
        self,
        graph: BidirectedGraph[type],
        dependee_injectable: type,
        dependent_contract: type,
        dependent_registrations: Iterable[InjectableRegistration]
    ) -> Iterable[type]:
        # Mark each implementation as a dependency. At this point,
        # it is possible only one of them will be needed by
        # this specific type. But we'll make sure all are initialized
        # as they may be needed later.
        dependent_types = []
        for dependent_registration in dependent_registrations:
            injectable_type = self.__get_injectable_type(
                dependent_contract,
                dependent_registration
            )

            graph.try_add_node(dependee_injectable)
            graph.try_add_edge(dependee_injectable, injectable_type)
            dependent_types.append(injectable_type)
            self.__log.debug(
                "Identified dependent injectable",
                dependee=dependee_injectable,
                dependent=injectable_type
            )

        return dependent_types

    def __get_injectable_type(
        self,
        dependent_contract: type,
        dependent_registration: InjectableRegistration
    ) -> type:
        match dependent_registration:
            case InjectableTypeRegistration():
                generic_type = self.__get_or_create_generic_type(
                    dependent_registration,
                    dependent_contract
                )
                return (
                    generic_type.closed_generic_type if generic_type
                    else dependent_registration.injectable_type
                )
            case InjectableInstanceRegistration():
                return type(dependent_registration.injectable_instance)
            case _:
                raise DependencyResolutionException(
                    type(dependent_registration),
                    "Unsupported type of injectable registration."
                )

    def __get_or_create_generic_type(
        self,
        registration: InjectableTypeRegistration,
        contract: type
    ) -> ClosedGenericTypeInfo | None:
        if not registration.is_generic:
            return None

        type_arguments = get_args(contract)
        if len(type_arguments) != 1:
            raise DependencyResolutionException(
                contract,
                "The generic contract must have one single generic type argument."
            )

        type_argument = type_arguments[0]
        generic_type_id = ClosedGenericTypeId(registration.injectable_type, type_argument)

        # Avoid creating the same type multiple times.
        if existing_type := self.__closed_generic_type_infos_by_id.get(generic_type_id):
            return existing_type

        closed_generic_type = types.new_class(
            f"{registration.injectable_type.__name__}Of{type_argument.__name__}",
            # Ignoring the type here, because the linter isn't aware this type is a Generic[T].
            (registration.injectable_type[type_argument],), # type: ignore
            None,
            lambda ns: ns.update({
                "generic_type_argument": property(lambda _: type_argument)
            })
        )
        generic_type_info = ClosedGenericTypeInfo(
            closed_generic_type=closed_generic_type,
            generic_type_argument=type_argument,
            origin_registration=registration
        )

        # Save the newly created type to avoid recreating it if it's needed later.
        self.__closed_generic_type_infos_by_id[generic_type_id] = generic_type_info
        self.__closed_generic_type_infos_by_type[closed_generic_type] = generic_type_info

        return generic_type_info



class InjectableRegistrationException(Exception):
    """Raised when registering an injectable fails."""

    def __init__(
        self,
        related_type: type | None = None,
        message: str | None = None
    ) -> None:
        """Initializes a new instance.

        :param related_type: An optional type related to the exception, defaults to None.
        :type related_type: type, optional
        :param message: An optional message that describes the problem, defaults to None.
        :type message: str, optional
        """

        super().__init__(message)
        self.__related_type = related_type

    @property
    def related_type(self) -> type | None:
        """Gets the type related to the exception.

        :return: The type related to the exception.
        :rtype: type | None
        """

        return self.__related_type



from dataclasses import dataclass

from kanata.catalogs import IInjectableCatalog
from kanata.models import ClosedGenericTypeId, ClosedGenericTypeInfo, IInstanceCollection

@dataclass(frozen=True, kw_only=True)
class ResolverContext:
    """Holds contextual information for resolvers."""

    catalog: IInjectableCatalog
    """The catalog of injectables."""

    closed_generic_types: dict[ClosedGenericTypeId, ClosedGenericTypeInfo]
    """A map of closed generic types available for injection."""

    instances: IInstanceCollection
    """The already resolved instances of injectables."""



from kanata.exceptions import ArgumentException
from kanata.graphs import BidirectedGraph, TNode
from kanata.graphs.exceptions import CyclicGraphException, DisconnectedSubGraphException

def topological_sort(graph: BidirectedGraph[TNode], start_node: TNode) -> tuple[TNode, ...]:
    """Produces a linear ordering of the specified graph's nodes
    such that for every directed edge uv from node u to node v, u comes before v in the ordering.

    :param graph: The graph whose nodes are to be sorted.
    :type graph: BidirectedGraph[TNode]
    :param start_node: The node used as the starting point for the ordering.
    :type start_node: TNode
    :raises ArgumentException: Raised when the specified start node is not in the graph.
    :raises DisconnectedSubGraphException: Raised when a disconnected sub-graph is detected within the graph.
    :return: The sorted nodes of the specified graph.
    :rtype: tuple[TNode, ...]
    """

    if start_node not in graph:
        raise ArgumentException(
            "start_node",
            start_node,
            f"The start node '{start_node}' is not in the graph."
        )

    visited_nodes: set[TNode] = set()
    unvisited_nodes: set[TNode] = graph.nodes
    sorted_nodes: list[TNode] = []
    __visit(graph, start_node, visited_nodes, set(), sorted_nodes)
    if visited_nodes != unvisited_nodes:
        raise DisconnectedSubGraphException(tuple(str(node) for node in unvisited_nodes))
    return tuple(sorted_nodes)

def __visit(
    graph: BidirectedGraph[TNode],
    node: TNode,
    visited_nodes: set[TNode],
    currently_visiting: set[TNode],
    sorted_nodes: list[TNode]
) -> None:
    if node in visited_nodes:
        return
    if node in currently_visiting:
        raise CyclicGraphException(
            currently_visiting,
            "A directed acyclic graph (DAG) is expected."
        )

    currently_visiting.add(node)
    for out_node in {edge.target for edge in graph.get_out_edges(node)}:
        __visit(graph, out_node, visited_nodes, currently_visiting, sorted_nodes)
    currently_visiting.remove(node)

    visited_nodes.add(node)
    sorted_nodes.append(node)



from dataclasses import dataclass

@dataclass
class DefaultResolverOptions:
    """Holds options for a default resolver."""

    suppress_captive_dependency_warnings: bool = False
    """Gets or sets whether to suppress warning logs about captive dependencies."""

    raise_on_captive_dependency: bool = True
    """Gets or sets whether to raise an exception when a captive dependency is detected."""



"""Utilities for discovering injectables."""

import importlib
import inspect
import pkgutil
from types import ModuleType
from typing import Any

import structlog

from .constants import LOGGER_NAME
from .models import InjectableRegistration

def find_injectables(module_name: str) -> tuple[InjectableRegistration, ...]:
    """Recursively discovers all injectables starting from the specified module.

    :param module_name: The name of the root module.
    :type module_name: str
    :return: The registration objects of the discovered injectables.
    :rtype: tuple[InjectableRegistration, ...]
    """

    logger = structlog.get_logger(logger_name=LOGGER_NAME)
    registrations: list[InjectableRegistration] = []
    module_names: list[str] = [module_name]
    while module_names:
        module_name = module_names.pop()
        logger.debug("Loading module", name=module_name)
        module = importlib.import_module(module_name)
        registrations.extend(__get_registrations(module, logger))

        if (path := getattr(module, "__path__", None)) is None:
            continue

        logger.debug("Walking path", path=path)
        for _, name, is_package in pkgutil.walk_packages(path):
            if not is_package:
                continue
            logger.debug("Found additional package", name=name, parent=module_name)
            module_names.append(f"{module_name}.{name}")

    return tuple(registrations)

def __get_registrations(module: ModuleType, logger: Any) -> tuple[InjectableRegistration, ...]:
    registrations: list[InjectableRegistration] = []
    for _, obj in inspect.getmembers(module, inspect.isclass):
        if local_registration := getattr(
            obj, InjectableRegistration.PROPERTY_NAME, None
        ):
            logger.debug("Found injectable", type=local_registration.injectable_type)
            registrations.append(local_registration)

    return tuple(registrations)



from typing import Protocol, TypeVar

from kanata.models import InjectableRegistration

T = TypeVar("T")

class IInjectableCatalog(Protocol):
    """Interface for a catalog of injectable registrations."""

    def get_registrations(self) -> tuple[InjectableRegistration, ...]:
        """Gets all the registrations available in the catalog.

        :return: The available registrations.
        :rtype: tuple[InjectableRegistration, ...]
        """
        ...

    def get_registrations_by_contract(
        self,
        contract: type) -> tuple[InjectableRegistration, ...]:
        """Gets the registrations associated to the specified contract.

        :param contract: The contract for which to get the registrations.
        :type contract: type
        :return: The registrations associated to the contract.
        :rtype: tuple[InjectableRegistration, ...]
        """
        ...

    def get_registration_by_injectable(
        self,
        injectable: type) -> InjectableRegistration | None:
        """Gets the registration associated to the specified injectable.

        :param injectable: The injectable for which to get the registration.
        :type injectable: type
        :return: If exists, the registration associated to the injectable.
        :rtype: InjectableRegistration | None
        """
        ...



from collections.abc import Generator
from typing import Any, Protocol

from .injectable_scope_type import InjectableScopeType

class IInstanceCollection(Protocol):
    """Interface for a resolved instance collection."""

    def get_instances_by_injectable(
        self,
        injectable_type: type,
        scope_type: InjectableScopeType | None = None
    ) -> Generator[Any, None, None]:
        """Gets all the resolved instances associated to the specified contract.

        :param injectable_type: The injectable type for which to get the instances.
        :type injectable_type: type
        :param scope_type: An optional scope for which to get the instances, defaults to None
        :type scope_type: InjectableScopeType | None, optional
        :yield: Resolved instances associated to the contract.
        :rtype: Generator[Any, None, None]
        """
        ...



from typing import Generic

from .tnode import TNode

class Edge(Generic[TNode]):
    """Defines an edge between two nodes of a graph."""

    def __init__(self, source: TNode, target: TNode) -> None:
        """Initializes a new instance.

        :param source: The source node of the edge.
        :type source: TNode
        :param target: The target node of the edge.
        :type target: TNode
        """

        super().__init__()
        self.__source: TNode = source
        self.__target: TNode = target

    @property
    def source(self) -> TNode:
        """Gets the source node of the edge.

        :return: The source node of the edge.
        :rtype: TNode
        """

        return self.__source

    @property
    def target(self) -> TNode:
        """Gets the target node of the edge.

        :return: The target node of the edge.
        :rtype: TNode
        """

        return self.__target

    def __eq__(self, other: object) -> bool:
        return (
            self.__source == other.source and self.__target == other.target
            if isinstance(other, Edge)
            else False
        )



from typing import Any

class ArgumentException(Exception):
    """Raised when a method is invoked with an invalid argument."""

    def __init__(self, name: str, value: Any, message: str | None = None) -> None:
        """Initializes a new instance.

        :param name: The name of the argument.
        :type name: str
        :param value: The actual value of the argument.
        :type value: Any
        :param message: An optional message that describes the problem, defaults to None.
        :type message: str, optional
        """

        super().__init__(message)
        self.__name: str = name
        self.__value: str = value

    @property
    def name(self) -> str:
        """Gets the name of the argument that caused the exception.

        :return: The name of the argument.
        :rtype: str
        """

        return self.__name

    @property
    def value(self) -> Any:
        """Gets the actual value of the argument that caused the exception.

        :return: The actual value of the argument.
        :rtype: Any
        """

        return self.__value



from collections.abc import Iterable
from typing import get_origin

from kanata.exceptions import InjectableRegistrationException
from kanata.models import (
    InjectableInstanceRegistration, InjectableRegistration, InjectableTypeRegistration
)
from kanata.utils import get_generic_type_parameters, get_or_add
from .iinjectable_catalog import IInjectableCatalog

class InjectableCatalog(IInjectableCatalog):
    """Provides access to information about the known injectables."""

    def __init__(self, registrations: Iterable[InjectableRegistration]) -> None:
        self.__registrations_by_contract: dict[type, list[InjectableRegistration]] = {}
        self.__registrations_by_injectable: dict[type, InjectableRegistration] = {}
        self.__build_registration_maps(registrations)

    def get_registrations(self) -> tuple[InjectableRegistration, ...]:
        return tuple(self.__registrations_by_injectable.values())

    def get_registrations_by_contract(
        self,
        contract: type
    ) -> tuple[InjectableRegistration, ...]:
        registrations = [*self.__registrations_by_contract.get(contract, ())]

        if (origin := get_origin(contract)) and get_generic_type_parameters(origin):
            registrations.extend(
                self.__registrations_by_contract.get(origin, ())
            )

        return tuple(registrations)

    def get_registration_by_injectable(
        self,
        injectable: type
    ) -> InjectableRegistration | None:
        return self.__registrations_by_injectable.get(injectable, None)

    def __build_registration_maps(self, registrations: Iterable[InjectableRegistration]) -> None:
        for registration in registrations:
            for contract_type in registration.contract_types:
                by_injectables: list[InjectableRegistration] = get_or_add(
                    self.__registrations_by_contract,
                    contract_type,
                    lambda _: [])
                by_injectables.append(registration)

            match registration:
                case InjectableTypeRegistration():
                    key = registration.injectable_type
                case InjectableInstanceRegistration():
                    key = type(registration.injectable_instance)
                case _:
                    raise InjectableRegistrationException(
                        type(registration),
                        "Unsupported type of injectable registration."
                    )

            self.__registrations_by_injectable[key] = registration



from __future__ import annotations

from collections.abc import Iterable
from typing import Any

from kanata import find_injectables
from kanata.exceptions import ArgumentException
from kanata.models import (
    InjectableInstanceRegistration, InjectableRegistration, InjectableScopeType,
    InjectableTypeRegistration
)
from kanata.utils.type_utils import get_generic_type_parameters
from .iinjectable_catalog import IInjectableCatalog
from .injectable_catalog import InjectableCatalog

class InjectableCatalogBuilder:
    """An injectable catalog builder."""

    def __init__(self) -> None:
        self.__registrations = list[InjectableRegistration]()

    def add_module(self, module_name: str) -> InjectableCatalogBuilder:
        """Discovers the injectable registrations from the module with the specified name.

        :param module_name: The name of the module to explore.
        :type module_name: str
        :return: The same instance of the builder.
        :rtype: InjectableCatalogBuilder
        """

        self.__registrations.extend(find_injectables(module_name))
        return self

    def register_instance(
        self,
        instance: Any,
        contract_types: Iterable[type]
    ) -> InjectableCatalogBuilder:
        """Registers the specified instance as an injectable.

        :param instance: The instance to be registered.
        :type instance: Any
        :param contract_types: The contracts by which to register the instance.
        :type contract_types: Iterable[type]
        :return: The same instance of the builder.
        :rtype: InjectableCatalogBuilder
        """

        # TODO https://github.com/PyCQA/pylint/issues/6550
        self.__registrations.append(
            InjectableInstanceRegistration( # pylint: disable=unexpected-keyword-arg
                contract_types=set(contract_types),
                injectable_instance=instance
            )
        )
        return self

    def register_type(
        self,
        injectable_type: type,
        contract_types: Iterable[type],
        scope_type: InjectableScopeType = InjectableScopeType.TRANSIENT
    ) -> InjectableCatalogBuilder:
        """Registers the specified type as an injectable.

        :param injectable_type: The type to be registered.
        :type injectable_type: type
        :param contract_types: The contracts by which to register the type.
        :type contract_types: Iterable[type]
        :param scope_type: The injectable scope, defaults to InjectableScopeType.TRANSIENT
        :type scope_type: InjectableScopeType, optional
        :return: The same instance of the builder.
        :rtype: InjectableCatalogBuilder
        """

        return self.__register_type(injectable_type, contract_types, False, scope_type)

    def register_generic(
        self,
        injectable_type: type,
        contract_types: Iterable[type],
        scope_type: InjectableScopeType = InjectableScopeType.TRANSIENT
    ) -> InjectableCatalogBuilder:
        """Registers the specified generic type as an injectable.

        :param injectable_type: The generic type to be registered.
        :type injectable_type: type
        :param contract_types: The contracts by which to register the generic type.
        :type contract_types: Iterable[type]
        :param scope_type: The injectable scope, defaults to InjectableScopeType.TRANSIENT
        :type scope_type: InjectableScopeType, optional
        :return: The same instance of the builder.
        :rtype: InjectableCatalogBuilder
        """

        InjectableCatalogBuilder.__validate_generic_type(injectable_type)
        for contract_type in contract_types:
            InjectableCatalogBuilder.__validate_generic_type(contract_type)

        return self.__register_type(injectable_type, contract_types, True, scope_type)

    def build(self) -> IInjectableCatalog:
        """Builds the injectable catalog.

        :return: The new instance of the injectable catalog.
        :rtype: IInjectableCatalog
        """

        return InjectableCatalog(self.__registrations)

    @staticmethod
    def __validate_generic_type(typ: type) -> None:
        if len(get_generic_type_parameters(typ)) != 1:
            raise ArgumentException(
                "typ",
                typ,
                "A generic injectable type must have a single generic type parameter."
            )

    def __register_type(
        self,
        injectable_type: type,
        contract_types: Iterable[type],
        is_generic: bool,
        scope_type: InjectableScopeType
    ) -> InjectableCatalogBuilder:
        # TODO https://github.com/PyCQA/pylint/issues/6550
        self.__registrations.append(
            InjectableTypeRegistration( # pylint: disable=unexpected-keyword-arg
                contract_types=set(contract_types),
                injectable_type=injectable_type,
                is_generic=is_generic,
                scope=scope_type
            )
        )
        return self



from .default_resolver import DefaultResolver
from .default_resolver_options import DefaultResolverOptions
from .iresolver import IResolver
from .resolver_base import ResolverBase
from .resolver_context import ResolverContext



from typing import Generic

from .edge import Edge
from .exceptions import DuplicateEdgeException, DuplicateNodeException
from .tnode import TNode

class Graph(Generic[TNode]):
    """A basic, non-directed graph that consists of nodes and edges between them."""

    def __init__(self, allow_parallel_edges: bool = False) -> None:
        """Initializes a new instance.

        :param allow_parallel_edges: Whether multiple edges between the same two nodes are allowed, defaults to False.
        :type allow_parallel_edges: bool, optional
        """

        self.__allow_parallel_edges: bool = allow_parallel_edges
        self.__nodes: set[TNode] = set()
        self.__edges: list[Edge[TNode]] = []

    @property
    def allow_parallel_edges(self) -> bool:
        """Gets whether multiple edges between the same two nodes are allowed.

        :return: Whether multiple edges between the same two nodes are allowed.
        :rtype: bool
        """

        return self.__allow_parallel_edges

    @property
    def nodes(self) -> set[TNode]:
        """Gets the set of nodes this graphs consist of.

        :return: The set of nodes this graphs consist of.
        :rtype: set[TNode]
        """

        return self.__nodes.copy()

    @property
    def edges(self) -> tuple[Edge[TNode], ...]:
        """Gets the collection of edges between the nodes.

        :return: The collection of edges between the nodes.
        :rtype: tuple[Edge[TNode], ...]
        """

        return tuple(self.__edges)

    def add_node(self, node: TNode) -> None:
        """Adds a node that hasn't been added to the graph yet.

        :param node: The node to be added.
        :type node: TNode
        :raises DuplicateNodeException: Raised when the node to be added is already part of the graph.
        """

        if node in self.__nodes:
            raise DuplicateNodeException(node, "The given node is already part of the graph.")
        self.__nodes.add(node)

    def try_add_node(self, node: TNode) -> bool:
        """Adds a new node to the graphi if it hasn't been added yet.

        :param node: The node to be added.
        :type node: TNode
        :return: True, if the node has been added; false, if it's already present.
        :rtype: bool
        """

        if node in self.__nodes:
            return False
        self.__nodes.add(node)
        return True

    def add_edge(self, source: TNode, target: TNode) -> None:
        """Adds an edge that may or may have not been added yet to the graph.

        :param source: The source node of the edge.
        :type source: TNode
        :param target: The target node of the edge.
        :type target: TNode
        :raises DuplicateEdgeException: Raised when parallel edges aren't allowed and an identical edge is already added.
        """

        if not self.try_add_edge(source, target):
            raise DuplicateEdgeException(source, target)

    def try_add_edge(self, source: TNode, target: TNode) -> bool:
        """Adds an edge to the graph between the specified nodes if a) such an edge doesn't exist yet, or b) parallel edges are allowed.

        :param source: The source node of the edge.
        :type source: TNode
        :param target: The target node of the edge.
        :type target: TNode
        :return: True, if the edge has been added; false, if it's already present.
        :rtype: bool
        """

        new_edge = Edge(source, target)
        if self.__allow_parallel_edges:
            self.__edges.append(new_edge)
            return True

        for edge in self.__edges:
            if self._is_same_edge(edge, new_edge):
                return False

        self.__edges.append(new_edge)
        return True

    def _is_same_edge(self, existing_edge: Edge, new_edge: Edge) -> bool:
        """Determines whether two edges are equal.

        Override this method in sub-classes to determine if two edges are equal. By default,
        two edges are equal if and only if their sources and targets are equal in any combination.

        :param existing_edge: The already existing edge.
        :type existing_edge: Edge
        :param new_edge: The new edge.
        :type new_edge: Edge
        :return: True, if the two edges are equal.
        :rtype: bool
        """

        return (
            (existing_edge.source == new_edge.source and existing_edge.target == new_edge.target)
            or (existing_edge.source == new_edge.target and existing_edge.target == new_edge.source)
        )

    def __contains__(self, item: object) -> bool:
        return item in self.__nodes



from collections.abc import Callable
from typing import TypeVar

from kanata.models import InjectableTypeRegistration
from kanata.utils import get_generic_type_parameters, get_or_add_attribute

T = TypeVar("T")

def injectable(contract_type: type) -> Callable[[type[T]], type[T]]:
    """Marks a class as an injectable.

    :param contract_type: The type of the contract by which an instance of the object is injectable.
    :type contract_type: type
    :return: A decorator that returns the same class that it decorates.
    :rtype: Callable[[type[T]], type[T]]
    """

    def decorator(wrapped_class: type[T]) -> type[T]:
        registration = get_or_add_attribute(
            wrapped_class,
            InjectableTypeRegistration.PROPERTY_NAME,
            lambda injectable_type=wrapped_class: __create_registration(injectable_type)
        )
        registration.contract_types.add(contract_type)
        return wrapped_class

    return decorator

def __create_registration(
    injectable_type: type
) -> InjectableTypeRegistration:
    # TODO https://github.com/PyCQA/pylint/issues/6550
    return InjectableTypeRegistration( # pylint: disable=unexpected-keyword-arg
        injectable_type=injectable_type,
        is_generic=bool(get_generic_type_parameters(injectable_type))
    )



"""Models."""

from .closed_generic_type_id import ClosedGenericTypeId
from .closed_generic_type_info import ClosedGenericTypeInfo
from .iinstance_collection import IInstanceCollection
from .injectable_instance_registration import InjectableInstanceRegistration
from .injectable_registration import InjectableRegistration
from .injectable_scope_type import InjectableScopeType
from .injectable_type_registration import InjectableTypeRegistration
from .instance_collection import InstanceCollection



from typing import NamedTuple

class ClosedGenericTypeId(NamedTuple):
    """A named tuple used to identify a dynamically created closed generic type."""

    origin_type: type
    """The type of the origin injectable from which the closed generic type was created."""

    generic_type_argument: type
    """The generic type argument of the closed generic type."""



from typing import Any

class DisconnectedSubGraphException(Exception):
    """Raised when a disconnected sub-graph is detected in a graph that doesn't allow it."""

    def __init__(self, nodes: tuple[Any, ...], *args: object) -> None:
        """Initializes a new instance.

        :param nodes: The list of nodes that form the disconnected sub-graph.
        :type nodes: tuple[Any, ...]
        """

        super().__init__(*args)
        self.__nodes: tuple[Any, ...] = nodes

    @property
    def nodes(self) -> tuple[Any, ...]:
        """Gets the list of nodes that form the disconnected sub-graph.

        :return: The list of nodes that form the disconnected sub-graph.
        :rtype: tuple[Any, ...]
        """

        return self.__nodes



from collections.abc import Callable
from typing import TypeVar

from kanata.models import InjectableScopeType, InjectableTypeRegistration
from kanata.utils import get_or_add_attribute

T = TypeVar("T")

def scope(scope_type: InjectableScopeType) -> Callable[[type[T]], type[T]]:
    """Specifies the scope type of an injectable class.

    :param scope_type: The lifetime scope of the created instances.
    :type scope_type: InjectableScopeType, optional
    :return: A decorator that returns the same class that it decorates.
    :rtype: Callable[[type[T]], type[T]]
    """
    def decorator(wrapped_class: type[T]) -> type[T]:
        registration = get_or_add_attribute(
            wrapped_class,
            InjectableTypeRegistration.PROPERTY_NAME,
            # TODO https://github.com/PyCQA/pylint/issues/6550
            lambda: InjectableTypeRegistration(injectable_type=wrapped_class) # pylint: disable=unexpected-keyword-arg
        )
        registration.scope = scope_type
        return wrapped_class

    return decorator



from enum import IntEnum

class InjectableScopeType(IntEnum):
    """Defines the valid types of an injectable's lifetime scope."""

    TRANSIENT = 0
    """Marks the injectable as transient,
    meaning a new instance is created for each dependency.
    """

    SINGLETON = 1
    """Marks the injectable as a singleton,
    meaning the same single instance is returned for
    each request from a lifetime scope and
    all of its child lifetime scopes.
    """

    SCOPED = 2
    """Marks the injectable as scoped,
    meaning a new instance is created for each lifetime scope,
    even the children of a parent lifetime scope (contrary to singletons).
    """



from dataclasses import dataclass, field
from typing import ClassVar

@dataclass(kw_only=True)
class InjectableRegistration:
    """Holds information about the registration of an injectable."""

    PROPERTY_NAME: ClassVar[str] = "kanata_injectable_registrations"
    """The name of the property used to hold the list
    of registration objects attached to an injectable."""

    contract_types: set[type] = field(default_factory=set)
    """Gets or sets the types of the contracts by which an instance of the object is injectable."""



from typing import Any

class DuplicateNodeException(Exception):
    """Raised when a duplicate node is detected in a graph."""

    def __init__(self, node: Any, *args: object) -> None:
        """Initializes a new instance.

        :param node: The node that causes the duplication.
        :type node: Any
        """

        super().__init__(*args)
        self.__node: Any = node

    @property
    def node(self) -> Any:
        """Gets the node that causes the duplication.

        :return: The node that causes the duplication.
        :rtype: Any
        """

        return self.__node



"""Utility methods for types."""

import inspect
from collections.abc import Callable, Generator
from typing import Generic, Protocol, TypeVar, get_args, get_origin

from kanata.exceptions import DependencyResolutionException

TAttribute = TypeVar("TAttribute")

def get_or_add_attribute(
    clazz: type,
    name: str,
    default_factory: Callable[[], TAttribute]
) -> TAttribute:
    """Gets the attribute with the specified name, if exists;
    otherwise adds it with the value determined by the specified value factory.

    :param clazz: The type of the class.
    :type clazz: type
    :param name: The name of the attribute.
    :type name: str
    :param default_factory: A value factory for the default value to add, if needed.
    :type default_factory: Callable[[], TAttribute]
    :return: The value of the attribute.
    :rtype: TAttribute
    """

    attribute: TAttribute | None = getattr(clazz, name, None)
    if attribute is None:
        attribute = default_factory()
        setattr(clazz, name, attribute)
    return attribute

def get_dependent_contracts(injectable: type) -> Generator[tuple[type, bool], None, None]:
    """Gets the types of contracts the specified injectable depends on.

    :param injectable: The injectable for which to get the contracts.
    :type injectable: type
    :raises DependencyResolutionException: Raised when the type is not a valid injectable.
    :yield: The type of the contract and whether multiple instances are allowed.
    :rtype: Generator[tuple[type, bool], None, None]
    """

    constructor = getattr(injectable, "__init__", None)
    if not constructor or not callable(constructor):
        raise DependencyResolutionException(injectable, "Invalid constructor.")
    signature = inspect.signature(constructor)
    for name, descriptor in signature.parameters.items():
        if name in ("self", "args", "kwargs"):
            continue
        if descriptor.annotation == inspect.Parameter.empty:
            raise DependencyResolutionException(
                injectable,
                f"The initializer of '{injectable}' is missing annotations."
            )
        yield __unpack_dependent_intf(descriptor.annotation)

def get_generic_type_parameters(typ: type) -> tuple[type, ...]:
    """Gets the generic type parameters of the specified type, if there are any.

    :param typ: The type to check.
    :type typ: type
    :return: The generic type parameters of the type.
    :rtype: tuple[type, ...]
    """

    if not (orig_bases := getattr(typ, "__orig_bases__", None)):
        return ()

    return tuple(set(
        generic_type_parameter
        for orig_base in orig_bases
        if get_origin(orig_base) in {Generic, Protocol}
        for generic_type_parameter in get_args(orig_base)
    ))

def __unpack_dependent_intf(contract: type) -> tuple[type, bool]:
    if (
        getattr(contract, "_is_protocol", False) # typing.Protocol
        or getattr(contract, "__abstractmethods__", None) # abc.ABCMeta
        or (origin := getattr(contract, "__origin__", None)) is None
    ):
        return (contract, False)

    if origin is not tuple:
        raise DependencyResolutionException(
            contract,
            f"Expected a tuple, but got {origin} for contract {contract}."
        )

    args = get_args(contract)
    if len(args) != 2 or args[-1] != Ellipsis:
        raise DependencyResolutionException(
            contract,
            "Expected a tuple with two arguments, the second being an ellipsis."
        )
    return (args[0], True)
